from mitama._extra import _Singleton
from mitama.app.router import Router
from mitama.app.method import group
from mitama.conf import get_from_project_dir
import sys
import os
import importlib
from watchdog.events import FileSystemEvent, PatternMatchingEventHandler
from watchdog.observers import Observer

class AppRegistry(_Singleton):
    '''稼働しているアプリのレジストリ

    サーバー内で稼働しているアプリのパスやパッケージ名が登録されているレジストリです。
    mitama.jsonを読み込んでアプリを起動するクラスでもあります。
    dictっぽくアプリの取得や配信の停止などが可能です。
    '''
    _map = dict()
    _server = None
    _router = None
    changed = False
    def __init__(self):
        super().__init__()
    def start_watch(self):
        observer = Observer()
        class Handler(PatternMatchingEventHandler):
            def on_event(self_):
                self.load_config()
            def on_created(self, event):
                self.on_event()
            def on_updated(self, event):
                self.on_event()
            def on_deleted(self, event):
                self.on_event()
        observer.schedule(Handler(), self.project_dir)
    def append(self, app):
        self._map.append(app)
    def __iter__(self):
        for app in self._map.values():
            yield app
    def __setitem__(self, path, app):
        self.changed = True
        self._map[path] = app
    def __getitem__(self, path):
        return self._map[path]
    def __delitem__(self, path):
        self.changed = True
        del self._map[path]
    def reset(self):
        '''アプリの一覧をリセットします'''
        self.changed = True
        self._map = dict()
    def load_config(self):
        '''アプリの一覧をmitama.jsonから読み込み、配信します'''
        config = get_from_project_dir()
        sys.path.append(str(config._project_dir))
        for app_name in config.apps:
            _app = config.apps[app_name]
            app_dir = config._project_dir / app_name
            if not app_dir.is_dir():
                os.mkdir(app_dir)
            if app_name not in sys.modules:
                init = importlib.__import__(app_name, fromlist=['AppBuilder'])
            else:
                init = importlib.reload(app_name)
            builder = init.AppBuilder()
            builder.set_package(app_name)
            builder.set_project_dir(config._project_dir / app_name)
            builder.set_project_root_dir(config._project_dir)
            builder.set_path(_app['path'])
            builder.set_name(app_name)
            app = builder.build()
            self[_app['path']] = app
    def router(self):
        '''アプリの情報に基づいてルーティングエンジンを生成します'''
        if self._router == None:
            router = Router()
            for k in self._map.keys():
                app = self._map[k]
                router.add_route(group(k, app))
            self._router = router
        elif self.changed:
            router = Router()
            for k in self._map.keys():
                app = self._map[k]
                router.add_route(group(k, app))
            self._router = router
            self.changed = False
        return self._router



import os
from pathlib import Path
from .controller import *
from .middleware import *
from mitama.app import App as BaseApp, Router, static_files
from mitama.app.method import *
from mitama.app.middlewares import SessionMiddleware
from .model import UpdateUserPermission, CreateUserPermission, DeleteUserPermission, CreateGroupPermission, UpdateGroupPermission, DeleteGroupPermission, Admin

import urllib

class App(BaseApp):
    name = "Mitama Portal"
    description = "Mitamaのアプリポータルです。他のアプリを確認できる他、配信の設定やグループの編集、ユーザーの招待ができます。"
    @property
    def view(self):
        view = super().view
        view.globals.update(
            user_create_permission = CreateUserPermission.is_accepted,
            user_update_permission = UpdateUserPermission.is_accepted,
            user_delete_permission = DeleteUserPermission.is_accepted,
            group_create_permission = CreateGroupPermission.is_accepted,
            group_update_permission = UpdateGroupPermission.is_accepted,
            group_delete_permission = DeleteGroupPermission.is_accepted,
            is_admin = Admin.is_accepted
        )
        return view
    @property
    def router(self):
        return Router([
            view('/static/<path:path>', static_files()),
            view('/setup', RegisterController, 'setup'),
            view('/signup', RegisterController, 'signup'),
            view('/login', SessionController, 'login'),
            Router([
                view('/', GroupsController, 'list'),
                view('/logout', SessionController, 'logout'),
                view('/users', UsersController, 'list'),
                view('/users/invite', UsersController, 'create'),
                view('/users/invite/<id>/delete', UsersController, 'cancel'),
                view('/users/<id>', UsersController, 'retrieve'),
                view('/users/<id>/settings', UsersController, 'update'),
                view('/users/<id>/delete', UsersController, 'delete'),
                view('/groups', GroupsController, 'list'),
                view('/groups/create', GroupsController, 'create'),
                view('/groups/<id>', GroupsController, 'retrieve'),
                post('/groups/<id>/append', GroupsController, 'append'),
                view('/groups/<id>/remove/<cid>', GroupsController, 'remove'),
                view('/groups/<id>/accept/<cid>/update', GroupsController, 'accept'),
                view('/groups/<id>/forbit/<cid>/update', GroupsController, 'forbit'),
                view('/groups/<id>/settings', GroupsController, 'update'),
                view('/groups/<id>/delete', GroupsController, 'delete'),
                view('/apps', AppsController, 'list'),
                view('/apps/settings', AppsController, 'update'),
            ], middlewares = [
                InitializeMiddleware,
                SessionMiddleware
            ])
        ])
    def delete_user(self, user):
        CreateUserPermission.forbit(user)
        UpdateUserPermission.forbit(user)
        DeleteUserPermission.forbit(user)
        CreateGroupPermission.forbit(user)
        UpdateGroupPermission.forbit(user)
        DeleteGroupPermission.forbit(user)
    def delete_group(self, group):
        CreateUserPermission.forbit(group)
        UpdateUserPermission.forbit(group)
        DeleteUserPermission.forbit(group)
        CreateGroupPermission.forbit(group)
        UpdateGroupPermission.forbit(group)
        DeleteGroupPermission.forbit(group)
    pass




from mitama.app.router import Route, GroupRoute

def view(path, handler, method = None):
    '''GET, POSTのルーティング先を指定

    ブラウザで見れる一般的なルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: Routeインスタンス
    '''
    return Route(['GET', 'POST'], path, handler, method)

def any(path, handler, method = None):
    '''メソッドを考慮しないルーティング先を指定

    メソッドに関係なくマッチするルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: Routeインスタンス
    '''
    return Route(['GET', 'POST', 'PATCH', 'PUT', 'DELETE', 'OPTION', 'HEAD'], path, handler, method)

def post(path, handler, method = None):
    '''POSTのルーティング先を指定

    POSTメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: Routeインスタンス
    '''
    return Route(['POST'], path, handler, method)

def patch(path, handler, method = None):
    '''PATCHのルーティング先を指定

    PATCHメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: Routeインスタンス
    '''
    return Route(['PATCH'], path, handler, method)

def head(path, handler, method = None):
    '''HEADのルーティング先を指定

    HEADメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: Routeインスタンス
    '''
    return Route(['HEAD'], path, handler, method)

def option(path, handler, method = None):
    '''OPTIONのルーティング先を指定

    OPTIONメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: Routeインスタンス
    '''
    return Route(['OPTION'], path, handler, method)

def put(path, handler, method = None):
    '''PUTのルーティング先を指定

    PUTメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: Routeインスタンス
    '''
    return Route(['PUT'], path, handler, method)

def get(path, handler, method = None):
    '''GETのルーティング先を指定

    GETメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: Routeインスタンス
    '''
    return Route(['GET'], path, handler, method)

def delete(path, handler, method = None):
    '''DELETEのルーティング先を指定

    DELETEメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: Routeインスタンス
    '''
    return Route(['DELETE'], path, handler, method)

def group(path, router):
    '''グループルーティング先を指定

    前方一致でルーティング先に中継するルーティング先を作成します。
    :param path: マッチするパス
    :param router: 渡すルーティングエンジン
    :return: GroupRouteインスタンス
    '''
    return GroupRoute(path, router)




#!/usr/bin/python

from mitama.nodes import User, Group, Relation
from mitama.db.types import Column, Integer, Node
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.schema import UniqueConstraint

class PermissionMixin(object):
    '''パーミッションのモデルの実装を支援します

    ホワイトリスト方式の許可システムを実装する上で役に立つ機能をまとめたクラスです。
    このクラスとBaseDatabase.Modelを継承したクラスを定義するとパーミッションシステムを実現できます。

    .. code-block:: python

        class SomePermission(PermissionMixin, db.Model):
            pass

    継承先のクラスで :samp:`target` プロパティを定義した場合、特定のものに対してのみ許可する仕様にすることができます。

    .. code-block:: python

        class SomePermission(PermissionMixin, db.Model):
            target = Column(User)

    targetがUser、またはGroupの場合、targetUpPropagate、 targetDownPropagateを指定すれば、targetに対しても伝播をチェックすることができます。

    :param _id: 固有のID
    :param node: 許可するUser、またはGroupのインスタンス
    :param targetUpPropagate: targetがUser、またはGroupの場合の上向き伝播
    :param targetDownPropagate: targetがUser、またはGroupの場合の下向き伝播
    :param upPropagate: 許可対象のUser、またはGroupの場合の上向き伝播
    :param downPropagate: 許可対象のUser、またはGroupの場合の下向き伝播
    '''
    @declared_attr
    def __tablename__(cls):
        return '__'+cls.__name__.lower()+'_permission'
    @declared_attr
    def __table_args__(cls):
        if hasattr(cls, 'target'):
            unique = UniqueConstraint('node', 'target', name='unique')
        else:
            unique = UniqueConstraint('node', name='unique')
        return (unique, )
    node = Column(Node)
    targetUpPropagate = False
    targetDownPropagate = False
    upPropagate = False
    downPropagate = False
    @classmethod
    def accept(cls, node, target = None):
        '''UserまたはGroupに許可します

        :param node: UserまたはGroupのインスタンス
        :param target: 許可対象
        '''
        perm = cls()
        perm.node = node
        if hasattr(cls, 'target') and target != None:
            perm.target = target
        perm.create()
    @classmethod
    def forbit(cls, node, target = None):
        '''UserまたはGroupの許可を取りやめます

        :param node: UserまたはGroupのインスタンス
        :param target: 許可対象
        '''
        if hasattr(cls, 'target'):
            perm = cls.query.filter(cls.node == node).filter(cls.target == target).first()
        else:
            perm = cls.query.filter(cls.node == node).first()
        if perm!=None:
            perm.delete()
    @classmethod
    def is_accepted(cls, node, target = None):
        '''UserまたはGroupが許可されているか確認します

        :param node: UserまたはGroupのインスタンス
        :param target: 許可対象
        '''
        perms = cls.query.filter(cls.node == node).all()
        for perm in perms:
            if perm.is_target(target) or perm.is_target(None):
                return True
        if node.__class__.__name__ == 'User':
            parents = node.parents()
            for group in parents:
                if cls.is_accepted(group, target):
                    return True
        for node_ in cls.query.all():
            if node_.node == None:
                continue
            if node_.upPropagate and not isinstance(node_.node, User) and node_.node.is_ancestor(node) and node_.is_target(target):
                return True
            if node_.downPropagate and not isinstance(node_.node, User) and node_.node.is_descendant(node) and node_.is_target(target):
                return True
        return False
    def is_target(self, target = None):
        if not hasattr(self, 'target'):
            return True
        if self.target == None:
            return True
        if self.target == target:
            return True
        if isinstance(target, User) or isinstance(target, Group):
            if self.targetUpPropagate:
                return self.target.is_ancestor(target)
            elif self.targetDownPropagate:
                return self.target.is_descendant(target)
            else:
                return self.target == target
        elif self.target == target:
            return True
        return False
    @classmethod
    def is_forbidden(cls, node, target = None):
        '''UserまたはGroupが許可されていないか確認します

        :param node: UserまたはGroupのインスタンス
        :param target: 許可対象
        '''
        if not hasattr(cls, 'target'):
            return not cls.is_accepted(node)
        else:
            return not cls.is_accepted(node, target)




class _Singleton:
    _instance = None
    def __new__(cls, *args, **kwargs):
        if cls._instance == None:
            cls._instance = super(_Singleton, cls).__new__(cls)
        return cls._instance


class _classproperty:
    def __init__(self, fget = None, doc = None):
        self.fget = fget
        self.__doc__ = self.fget.__doc__ if doc == None else doc
    def __get__(self, obj, cls=None):
        if cls is None:
            cls = type(obj)
        return self.fget(cls)




from mitama.db import _CoreDatabase
from mitama.db.driver.sqlite3 import get_test_engine
from sqlalchemy import *
from sqlalchemy.orm import *

def create_test_db():
    engine = get_test_engine()
    return engine

def create_session(engine):
    session = scoped_session(
        sessionmaker(
            autocommit = False,
            autoflush = False,
            bind = engine
        )
    )
    return session

def test_nodes():
    engine = create_test_db()
    db = _CoreDatabase(engine)
    from mitama.nodes import User, Group
    db.create_all()
    session = create_session(engine)
    session.execute('insert into mitama_user (_id, name, screen_name, password) values (123, "test_user", "test_user_screen", "test_user_password")')
    session.execute('insert into mitama_group (_id, name, screen_name) values (456, "test_group", "test_group_screen")')
    test_user = User.query.first()
    test_group = Group.query.first()
    assert test_user._id == 123
    assert test_user.name == 'test_user'
    assert test_user.screen_name == 'test_user_screen'
    assert test_user.password == 'test_user_password'
    assert test_group._id == 456
    assert test_group.name == 'test_group'
    assert test_group.screen_name == 'test_group_screen'

def test_relation():
    db = _CoreDatabase()
    from mitama.nodes import User,Group,Relation
    db.create_all()
    test_user = User()
    test_user.name = 'test_user_'
    test_user.screen_name = 'test_user_screen_'
    test_user.password = 'test_pass_'
    db.session.add(test_user)
    test_group = Group()
    test_group.name = 'test_group_'
    test_group.screen_name = 'test_group_screen_'
    test_group2 = Group()
    test_group2.name = 'test_group_2'
    test_group2.screen_name = 'test_group_screen_2'
    db.session.add(test_group)
    db.session.add(test_group2)
    db.session.commit()
    test_group.append_all([
        test_user,
        test_group2
    ])
    test_children = test_group.children()
    assert test_children[0] == test_user
    assert test_children[1] == test_group2
    test_group.remove(test_user)
    test_group.remove(test_group2)
    test_children = test_group.children()
    assert len(test_children) == 0

def test_relation():
    db = _CoreDatabase()
    from mitama.nodes import User,Group,Relation
    db.create_all()
    test_user = User()
    test_user.name = 'test_user_'
    test_user.screen_name = 'test_user_screen_'
    test_user.password = 'test_pass_'
    db.session.add(test_user)
    test_group = Group()
    test_group.name = 'test_group_'
    test_group.screen_name = 'test_group_screen_'
    test_group2 = Group()
    test_group2.name = 'test_group_2'
    test_group2.screen_name = 'test_group_screen_2'
    db.session.add(test_group)
    db.session.add(test_group2)
    db.session.commit()
    test_group.append_all([
        test_user,
        test_group2
    ])
    test_children = test_group.children()
    assert test_children[0] == test_user
    assert test_children[1] == test_group2
    test_group.remove(test_user)
    test_group.remove(test_group2)
    test_children = test_group.children()
    assert len(test_children) == 0

def test_crud():
    db = _CoreDatabase()
    from mitama.nodes import User,Group
    test_user = User()
    test_user.name = 'test_user__'
    test_user.screen_name = 'test_user_screen__'
    test_user.password = 'test_pass__'
    test_user.create()
    searched = User.query.filter(User.screen_name == 'test_user_screen__').first()
    assert test_user == searched
    test_user.name = 'test_user_changed'
    test_user.update()
    searched = User.query.filter(User.screen_name == 'test_user_screen__').first()
    assert test_user == searched
    test_user.delete()
    searched = User.query.filter(User.screen_name == 'test_user_screen__').all()
    assert len(searched) == 0





from mitama.db import _CoreDatabase
from mitama.db.driver.sqlite3 import get_test_engine
from sqlalchemy import *
from sqlalchemy.orm import *
from datetime import datetime

def create_test_db():
    engine = get_test_engine()
    return engine

def create_session(engine):
    session = scoped_session(
        sessionmaker(
            autocommit = False,
            autoflush = False,
            bind = engine
        )
    )
    return session

def test_model():
    engine = create_test_db()
    db = _CoreDatabase(engine)
    class Todo(db.Model):
        __tablename__ = 'test_todo'
        title = Column(String(255))
        description = Column(String(255))
        datetime = Column(DateTime)
    db.create_all()
    session = create_session(engine)
    session.execute('insert into test_todo (_id, title, description, datetime) values (123, "test todo", "this is the test", datetime("2020-08-04 12:00:00"))')
    test_todo = Todo.query.first()
    assert test_todo._id == 123
    assert test_todo.title == 'test todo'
    assert test_todo.description == 'this is the test'
    assert test_todo.datetime == datetime(2020, 8, 4, 12)



import inspect
import os

class Builder(object):
    '''Appにメタ情報を入力して生成するビルダー

    プロジェクトディレクトリのパスやインストール先、プロジェクト内のアプリ用ディレクトリのパスをAppに設定し、インスタンスを返却します。
    特にこれを弄るケースは想定していませんが、独自の挙動を付けたかったら継承して作っても良いかもしれません。
    アプリのパッケージ直下のAppBuilderが起動されるので、:file:`__init__.py` に :samp:`class AppBuilder(Builder)` を定義してください。
    '''
    app = None
    def __init__(self):
        self.data = {}
        pass
    def set_path(self, path):
        self.data['path'] = path
    def set_name(self, name):
        self.data['name'] = name
    def set_package(self, package):
        self.data['package'] = package
    def set_project_dir(self, path):
        self.data['project_dir'] = path
    def set_project_root_dir(self, path):
        self.data['project_root_dir'] = path
    def build(self):
        install_dir = os.path.dirname(inspect.getfile(self.__class__))
        self.data['install_dir'] = install_dir
        return self.app(**self.data)



from mitama.app import BaseMetadata
from mitama.db import get_app_engine, Database as BaseDatabase

class Metadata(BaseMetadata):
    pass

class Database(BaseDatabase):
    def __init__(self, engine = None):
        super().__init__()
        meta = Metadata()
        if self.engine == None:
            if engine == None:
                self.set_engine(get_app_engine(meta.name))
            else:
                self.set_engine(engine)

def init_app(name):
    meta = Metadata()
    meta.name = name



#!/usr/bin/python
'''データベースのドライバ

    * sqlalchemyのデータベースエンジンを返す関数が入ってる
    * ぶっちゃけ、sqlite以外に対応する気はあんまりない
    * もしかしたらユーザーとグループの管理だけはmysqlとかに対応するかもしれない
    * アプリは廃棄可能性を保持したいので、絶対にファイルベースが良い
    * get_engine関数がユーザー管理用データベースのエンジンを返す
    * get_test_engine関数がユーザー管理用データベースのテスト用エンジンを返す
    * get_app_engine関数がアプリのデータベースのエンジンを返す(sqlite限定)
'''



#!/usr/bin/python
from yarl import URL
from jinja2 import Environment, FileSystemLoader
from pathlib import Path
import magic
import os
from base64 import b64encode
from mitama.noimage import load_noimage_app
from mitama.hook import HookRegistry
from mitama.http import Request, Response

def dataurl(blob):
    f = magic.Magic(mime = True, uncompress = True)
    mime = f.from_buffer(blob)
    return 'data:'+mime+';base64,'+b64encode(blob).decode()

class App:
    template_dir = 'templates'
    description = ""
    name = ""
    @property
    def icon(self):
        return load_noimage_app()
    def __init__(self, **kwargs):
        self.screen_name = kwargs['name']
        self.path = kwargs['path']
        self.package = kwargs['package']
        self.project_dir = Path(kwargs['project_dir']) if 'project_dir' in kwargs else None
        self.project_root_dir = Path(kwargs['project_root_dir']) if 'project_dir' in kwargs else None
        self.install_dir = Path(kwargs['install_dir']) if 'project_dir' in kwargs else Path(os.path.dirname(__file__)) / '../http/'
        self.router._app = self
        hook_registry = HookRegistry()
        if hasattr(self, 'create_user'):
            hook_registry.add_create_user_hook(self.create_user)
        if hasattr(self, 'create_group'):
            hook_registry.add_create_group_hook(self.create_group)
        if hasattr(self, 'update_user'):
            hook_registry.add_update_user_hook(self.update_user)
        if hasattr(self, 'update_group'):
            hook_registry.add_update_group_hook(self.update_group)
        if hasattr(self, 'delete_user'):
            hook_registry.add_delete_user_hook(self.delete_user)
        if hasattr(self, 'delete_group'):
            hook_registry.add_delete_group_hook(self.delete_group)
    def __call__(self, request):
        if not isinstance(request, Request):
            request = Request.from_request(request)
        result = self.router.match(request)
        if result:
            request, handle, method = result
            return handle(request)
        else:
            return self.error(request, 404)
    def set_middleware(self, middlewares):
        self.app.middlewares.extend(middlewares)
    def convert_fullurl(self, req, url):
        scheme = req.scheme
        hostname = req.host
        path = self.path
        if path[0] != '/':
            path = '/' + path
        if path[-1] == '/':
            path = path[0:-2]
        if url[0] != '/':
            url = '/' + url
        return scheme + "://" + hostname + path + url
    def convert_url(self, url):
        path = self.path
        if path[0] != '/':
            path = '/' + path
        if path[-1] == '/':
            path = path[0:-2]
        if url[0] != '/':
            url = '/' + url
        return path + url
    def revert_url(self, url):
        path = self.path
        if path[0] != '/':
            path = '/' + path
        if path[-1] == '/':
            path = path[0:-2]
        url = str(url.path)
        url = URL(url[len(path):])
        return url
    @property
    def view(self):
        self._view= Environment(
            loader = FileSystemLoader(self.install_dir / self.template_dir)
        )
        def filter_user(arg):
            return [user for user in arg if user.__class__.__name__ == "User"]
        def filter_group(arg):
            return [group for group in arg if group.__class__.__name__ == "Group"]
        self._view.filters["user"] = filter_user
        self._view.filters["group"] = filter_group
        self._view.globals.update(
            url = self.convert_url,
            fullurl = self.convert_fullurl,
            dataurl = dataurl
        )
        return self._view
    def error(self, request, code):
        template = self.view.get_template(str(code) + '.html')
        return Response.render(template)
    def match(self, request):
        result = self.router.match(request)
        if result == False:
            return False
        else:
            request, handle, method = self.router.match(request)
            def _handle(request):
                request.app = self
                return handle(request)
            return request, _handle, method

def _session_middleware():
    from mitama.http.session import EncryptedCookieStorage
    from cryptography import fernet
    from mitama.app import Middleware
    import base64
    class SessionMiddleware(Middleware):
        fernet_key = fernet.Fernet.generate_key()
        def __init__(self):
            secret_key = base64.urlsafe_b64decode(self.fernet_key)
            cookie_storage = EncryptedCookieStorage(secret_key)
            self.storage = cookie_storage
        def process(self, request, handler):
            request['mitama_session_storage'] = self.storage
            raise_response = False
            response = handler(request)
            if not isinstance(response, Response):
                return response
            session = request.get('mitama_session')
            if session is not None:
                if session._changed:
                    self.storage.save_session(request, response, session)
            if raise_response:
                raise response
            return response
    return SessionMiddleware

class _MainApp(App):
    def __init__(self, app_registry):
        self.app_registry = app_registry
        self._router = None
        super().__init__(name='_mitama', path='/', package = '_mitama')
    @property
    def router(self):
        if self._router == None or self.app_registry.changed:
            router = self.app_registry.router()
            router.add_middleware(_session_middleware())
            self._router = router
        return self._router



from jinja2 import *
from .app import App, _MainApp
from .router import Router
from .builder import Builder
from .registry import AppRegistry
from mitama.http import Request, Response
from pathlib import Path
from mimetypes import add_type, guess_type
from abc import ABCMeta, abstractmethod
import os

add_type('application/json', '.map')

class Controller():
    '''MVCのControllerの基底クラス

    メソッドを記述すると、それをリクエスト処理に利用できる。
    ルーティング時にメソッドを特に指定しない場合はhandleメソッドが実行される。
    :param app: Controllerを起動するAppのインスタンスの参照
    :param view: Controllerが利用するJinja2のEnvironmentインスタンス
    '''
    app = None
    view = None
    def __init__(self, app):
        self.app = app
        self.view = app.view
    def handle(self, request: Request):
        '''リクエストハンドラ

        ルーティング時に特にメソッド名を指定しない場合にはこのメソッドが起動する。
        独自にメソッドを定義する場合にも、このメソッドと同じインターフェースを実装しなければならない。
        :param request: mitama.http.Requestのインスタンス
        :return: mitama.http.Responseのインスタンス
        '''
        pass
    def __call__(self, request, method = None):
        if method == None:
            return self.handle(request)
        else:
            func = getattr(self, method)
            return func(request)

class Middleware(metaclass = ABCMeta):
    '''Requestを加工するMiddlewareの抽象クラス

    processメソッドによって受け取ったRequestを変更し、handler（次のMiddleware、またはControllerのメソッド）に受け渡す。
    :param app: Middlewareを起動するAppのインスタンスの参照
    :param view: Middlewareが利用するJinja2のEnvironmentインスタンス
    '''
    def __init__(self, app):
        self.app = app
        self.view = app.view
    @abstractmethod
    def process(self, request: Request, handler):
        '''Middlewareのメイン処理

        Middlewareは必ずこのメソッドを実装しなければならない。
        :param request: mitama.http.Requestのインスタンス
        :param handler: requestを引数に受け取る関数（Middleware.process、またはControllerのリクエストハンドラ）
        '''
        pass
    def __call__(self, request, handler):
        return self.process(request, handler)

def static_files(*paths):
    paths_ = list(paths)
    class StaticFileController(Controller):
        '''静的ファイルを配信するController

        デフォルトではアプリのパッケージ内の :file:`static/` の中身を配信する。
        '''
        paths = paths_
        def __init__(self, app):
            super().__init__(app)
            app_mod_dir = Path(os.path.dirname(__file__))
            self.view = Environment(
                loader = FileSystemLoader([
                    app_mod_dir / 'templates',
                    app_mod_dir / '../http/templates'
                ])
            )
            if len(self.paths) == 0:
                self.paths.append(self.app.install_dir / 'static')
        def handle(self, req: Request):
            for path in self.paths:
                filename = path / req.params['path']
                if filename.is_file():
                    mime = guess_type(str(filename)) or ('application/octet-stream', )
                    with open(filename, 'rb') as f:
                        return Response(body = f.read(), content_type = mime[0])
            for path in self.paths:
                filename = path / '404.html'
                if filename.is_file():
                    with open(filename) as f:
                        return Response(text = f.read(), status = 404, headers = {
                            'content-type': 'text/html'
                        })
            template = self.view.get_template('404.html')
            return Response.render(template, status = 404)
    return StaticFileController




from mitama.db.types import *
from mitama.nodes import User, Group
from mitama.db import BaseDatabase
from base64 import b64encode
import magic
from mitama.permission import PermissionMixin

class Database(BaseDatabase):
    pass

db = Database()

class Invite(db.Model):
    icon = Column(LargeBinary)
    screen_name = Column(String)
    name = Column(String)
    token = Column(String, unique = True)
    editable = Column(Boolean)
    def icon_to_dataurl(self):
        f = magic.Magic(mime = True, uncompress = True)
        mime = f.from_buffer(self.icon)
        return 'data:'+mime+';base64,'+b64encode(self.icon).decode()

class CreateUserPermission(PermissionMixin, db.Model):
    upPropagate = True
    pass

class UpdateUserPermission(PermissionMixin, db.Model):
    upPropagate = True
    targetDownPropagate = True
    target = Column(User.type, nullable = True)
    pass

class DeleteUserPermission(PermissionMixin, db.Model):
    upPropagate = True
    pass

class CreateGroupPermission(PermissionMixin, db.Model):
    upPropagate = True
    pass

class UpdateGroupPermission(PermissionMixin, db.Model):
    upPropagate = True
    targetDownPropagate = True
    target = Column(Group.type, nullable = True)
    pass

class DeleteGroupPermission(PermissionMixin, db.Model):
    upPropagate = True
    pass

class Admin(PermissionMixin, db.Model):
    upPropagate = True
    pass

db.create_all()



from mitama.db.driver.sqlite3 import get_test_engine 
from mitama.db import _CoreDatabase
from mitama.db import types

engine = get_test_engine()
db = _CoreDatabase(engine)
from mitama.nodes import User, Group, Relation
from mitama.permission import PermissionMixin

class HogePermission(PermissionMixin, db.Model):
    upPropagate = True
    pass

class PiyoPermission(PermissionMixin, db.Model):
    pass

class FugaPermission(PermissionMixin, db.Model):
    downPropagate = True
    pass

class HogeTargetedPermission(PermissionMixin, db.Model):
    target = types.Column(types.Group)
    targetUpPropagate = True
    pass

class PiyoTargetedPermission(PermissionMixin, db.Model):
    target = types.Column(types.Group)
    pass

class FugaTargetedPermission(PermissionMixin, db.Model):
    target = types.Column(types.Group)
    targetDownPropagate = True
    pass

class HogePiyoTargetedPermission(PermissionMixin, db.Model):
    target = types.Column(types.Group, nullable = True)
    pass

db.create_all()

u = list()
for i in range(5):
    u_ = User()
    u_.name = 'test'+str(i)
    u_.screen_name = 'test'+str(i)
    u_.password = ''
    u_.create()
    u.append(u_)

g = list()
for i in range(5):
    g_ = Group()
    g_.name = 'test'+str(i)
    g_.screen_name = 'test'+str(i)
    g_.create()
    g_.append(u[i])
    g.append(g_)
    if i > 0:
        g[i-1].append(g[i])

def test_up_propagate():
    HogePermission.accept(g[2])
    assert HogePermission.is_accepted(g[0])
    assert HogePermission.is_accepted(g[1])
    assert HogePermission.is_accepted(g[2])
    assert HogePermission.is_accepted(u[0])
    assert HogePermission.is_accepted(u[1])
    assert HogePermission.is_accepted(u[2])
    assert HogePermission.is_forbidden(g[3])
    assert HogePermission.is_forbidden(g[4])
    assert HogePermission.is_forbidden(u[3])
    assert HogePermission.is_forbidden(u[4])

def test_no_propagate():
    PiyoPermission.accept(g[2])
    assert PiyoPermission.is_accepted(g[2])
    assert PiyoPermission.is_accepted(u[2])
    assert PiyoPermission.is_forbidden(g[0])
    assert PiyoPermission.is_forbidden(g[1])
    assert PiyoPermission.is_forbidden(u[0])
    assert PiyoPermission.is_forbidden(u[1])
    assert PiyoPermission.is_forbidden(g[3])
    assert PiyoPermission.is_forbidden(g[4])
    assert PiyoPermission.is_forbidden(u[3])
    assert PiyoPermission.is_forbidden(u[4])

def test_down_propagate():
    FugaPermission.accept(g[2])
    assert FugaPermission.is_accepted(g[3])
    assert FugaPermission.is_accepted(g[4])
    assert FugaPermission.is_accepted(g[2])
    assert FugaPermission.is_accepted(u[3])
    assert FugaPermission.is_accepted(u[4])
    assert FugaPermission.is_accepted(u[2])
    assert FugaPermission.is_forbidden(g[0])
    assert FugaPermission.is_forbidden(g[1])
    assert FugaPermission.is_forbidden(u[0])
    assert FugaPermission.is_forbidden(u[1])

def test_target_up_propagate():
    HogeTargetedPermission.accept(u[2], g[2])
    assert HogeTargetedPermission.is_accepted(u[2], g[0])
    assert HogeTargetedPermission.is_accepted(u[2], g[1])
    assert HogeTargetedPermission.is_accepted(u[2], u[0])
    assert HogeTargetedPermission.is_accepted(u[2], u[1])
    assert HogeTargetedPermission.is_accepted(u[2], g[2])
    assert HogeTargetedPermission.is_forbidden(u[2], u[3])
    assert HogeTargetedPermission.is_forbidden(u[2], u[4])
    assert HogeTargetedPermission.is_forbidden(u[2], g[3])
    assert HogeTargetedPermission.is_forbidden(u[2], g[4])

def test_target_propagate():
    PiyoTargetedPermission.accept(u[2], g[2])
    assert PiyoTargetedPermission.is_forbidden(u[2], g[0])
    assert PiyoTargetedPermission.is_forbidden(u[2], g[1])
    assert PiyoTargetedPermission.is_forbidden(u[2], u[0])
    assert PiyoTargetedPermission.is_forbidden(u[2], u[1])
    assert PiyoTargetedPermission.is_accepted(u[2], g[2])
    assert PiyoTargetedPermission.is_forbidden(u[2], u[3])
    assert PiyoTargetedPermission.is_forbidden(u[2], u[4])
    assert PiyoTargetedPermission.is_forbidden(u[2], g[3])
    assert PiyoTargetedPermission.is_forbidden(u[2], g[4])

def test_target_down_propagate():
    FugaTargetedPermission.accept(u[2], g[2])
    assert FugaTargetedPermission.is_forbidden(u[2], g[0])
    assert FugaTargetedPermission.is_forbidden(u[2], g[1])
    assert FugaTargetedPermission.is_forbidden(u[2], u[0])
    assert FugaTargetedPermission.is_forbidden(u[2], u[1])
    assert FugaTargetedPermission.is_accepted(u[2], g[2])
    assert FugaTargetedPermission.is_accepted(u[2], u[3])
    assert FugaTargetedPermission.is_accepted(u[2], u[4])
    assert FugaTargetedPermission.is_accepted(u[2], g[3])
    assert FugaTargetedPermission.is_accepted(u[2], g[4])

def test_target_none_propagate():
    HogePiyoTargetedPermission.accept(u[3])
    assert HogePiyoTargetedPermission.is_accepted(u[3], g[0])
    assert HogePiyoTargetedPermission.is_accepted(u[3], u[0])
    assert HogePiyoTargetedPermission.is_accepted(u[3], u[1])
    assert HogePiyoTargetedPermission.is_accepted(u[3], g[2])
    assert HogePiyoTargetedPermission.is_accepted(u[3], u[2])
    assert HogePiyoTargetedPermission.is_accepted(u[3], u[4])
    assert HogePiyoTargetedPermission.is_accepted(u[3], g[3])
    assert HogePiyoTargetedPermission.is_accepted(u[3], g[4])
    assert HogePiyoTargetedPermission.is_forbidden(u[4], g[0])
    assert HogePiyoTargetedPermission.is_forbidden(u[4], g[1])
    assert HogePiyoTargetedPermission.is_forbidden(u[4], u[0])
    assert HogePiyoTargetedPermission.is_forbidden(u[4], u[1])
    assert HogePiyoTargetedPermission.is_forbidden(u[4], g[2])
    assert HogePiyoTargetedPermission.is_forbidden(u[4], u[2])
    assert HogePiyoTargetedPermission.is_forbidden(u[4], u[4])
    assert HogePiyoTargetedPermission.is_forbidden(u[4], g[3])
    assert HogePiyoTargetedPermission.is_forbidden(u[4], g[4])





from mitama.http import Response
from mitama.auth import AuthorizationError, check_jwt
from mitama.app import Middleware
import urllib

class SessionMiddleware(Middleware):
    '''ログイン判定ミドルウェア

    ログインしていないユーザーがアクセスした場合、/login?redirect_to=<URL>にリダイレクトします。
    '''
    def process(self, request, handler):
        sess = request.session()
        try:
            if 'jwt_token' in sess:
                request.user = check_jwt(sess['jwt_token'])
            else:
                return Response.redirect('/login?redirect_to='+urllib.parse.quote(str(request.url), safe=''))
        except Exception as err:
            return Response.redirect('/login?redirect_to='+urllib.parse.quote(str(request.url), safe=''))
        return handler(request)




from mitama.db.driver.sqlite3 import get_test_engine 
from mitama.db import _CoreDatabase

engine = get_test_engine()
db = _CoreDatabase(engine)
from mitama.nodes import User, Group, Relation
db.create_all()

u = list()
for i in range(5):
    u_ = User()
    u_.name = 'test'+str(i)
    u_.screen_name = 'test'+str(i)
    u_.password = ''
    u_.create()
    u.append(u_)

g = list()
for i in range(5):
    g_ = Group()
    g_.name = 'test'+str(i)
    g_.screen_name = 'test'+str(i)
    g_.create()
    g_.append(u[i])
    g.append(g_)
    if i > 0:
        g[i-1].append(g[i])

def test_ancestor():
    assert g[4].is_ancestor(g[0])
    assert g[4].is_ancestor(g[1])
    assert g[4].is_ancestor(g[2])
    assert g[4].is_ancestor(g[3])
    assert not g[4].is_ancestor(g[4])
    assert not g[0].is_ancestor(g[4])
    assert not g[1].is_ancestor(g[4])
    assert not g[2].is_ancestor(g[4])
    assert not g[3].is_ancestor(g[4])

def test_descendant():
    assert g[0].is_descendant(g[4])
    assert g[1].is_descendant(g[4])
    assert g[2].is_descendant(g[4])
    assert g[3].is_descendant(g[4])
    assert not g[4].is_descendant(g[4])
    assert not g[4].is_descendant(g[0])
    assert not g[4].is_descendant(g[1])
    assert not g[4].is_descendant(g[2])
    assert not g[4].is_descendant(g[3])



from mitama.http import Request, Response
import inspect
import copy
import re

class RoutingError(Exception):
    pass


class Router():
    '''ルーティングエンジン

    手軽に実装できて必要最低限なものを目指したので、遅かったりして嫌いだったら無理にこれを使う必要はありません。
    自前のものを適用したい場合は、とりあえず:samp:`hoge.match(Request): -> Response` といったインターフェースを実装したメソッドを作ってください。
    routesの中にRouterインスタンスを指定することもできます。
    '''
    _app = None
    def __init__(self, routes = [], middlewares = [], prefix = ''):
        '''初期化処理

        :param routes: Router、またはRouteのリスト
        :param middlewares: Middlewareのリスト
        '''
        self._parent = None
        self.routes = list()
        self.middlewares = list()
        self.add_routes(routes)
        self.add_middlewares(middlewares)
        self.i = 0
    def add_route(self, route):
        '''ルーティング先を追加します

        mitama.app.methodの関数で生成したRouteインスタンスを与えてください。
        :param route: Routeインスタンス
        '''
        if isinstance(route, Router):
            route._parent = self
        self.routes.append(route)
    def add_routes(self, routes):
        '''複数のルーティング先を追加します

        mitama.app.methodの関数で生成したRouteインスタンスを与えてください。
        :param routes: Routeインスタンスのリスト
        '''
        for route in routes:
            self.add_route(route)
    def add_middleware(self, middleware):
        '''ミドルウェアを登録します

        Middlewareクラスを与えると、このルーターのインスタンス内でマッチした場合にミドルウェアが順番に起動します。
        :param middleware: Middlewareのインスタンス
        '''
        self.middlewares.append(middleware)
    def add_middlewares(self, middlewares):
        '''複数のミドルウェアを登録します

        Middlewareクラスのリストを与えると、このルーターのインスタンス内でマッチした場合にミドルウェアが順番に起動します。
        :param middlewares: Middlewareのインスタンスのリスト
        '''
        for middleware in middlewares:
            self.add_middleware(middleware)
    def clone(self):
        return Router(routes = copy.copy(self.routes))
    def match(self, request):
        method = request.method
        path = request.subpath if hasattr(request, 'subpath') else request.path
        paths_to_check = [path]
        paths_to_check.append(re.sub('//+', '/', path))
        if not request.path.endswith('/'):
            paths_to_check.append(path + '/')
        paths_to_check.append(
            re.sub('//+', '/', path + '/'))
        if path.endswith('/'):
            merged_slashes = re.sub('//+', '/', path)
            paths_to_check.append(merged_slashes[:-1])

        for path in paths_to_check:
            for route in self.routes:
                request.subpath = path
                result = route.match(request)
                if result != False:
                    request, result, method = result
                    def get_response_handler(result, method):
                        i = 0
                        def handle(request):
                            nonlocal i
                            nonlocal result
                            if inspect.isclass(result):
                                result = result(request.app)
                                if method is not None:
                                    inst = result
                                    def result(request):
                                        return inst(request, method)
                            if i>=len(self.middlewares) or len(self.middlewares) == 0:
                                if callable(result):
                                    return result(request)
                                else:
                                    raise RoutingError('Unsupported interface object. Only callables and Controller instances are supported.')
                            else:
                                if hasattr(request, 'app'):
                                    middleware = self.middlewares[i](request.app)
                                else:
                                    middleware = self.middlewares[i]()
                                i += 1
                                return middleware.process(request, handle)
                        return handle
                    handler = get_response_handler(result, method)
                    return request, handler, None
        return False

class Route():
    def __init__(self, methods, path, handler, method_name):
        self.methods = methods
        self.path = Path(path)
        self.handler = handler
        self.method_name = method_name
        pass
    def match(self, request):
        method = request.method
        path = request.subpath if hasattr(request, 'subpath') else request.path
        args = self.path.match(path)
        if method in self.methods and args != False:
            request.params = args
            return request, self.handler, self.method_name
        else:
            return False

class GroupRoute():
    def __init__(self, path, router):
        self.path = path
        self.router = router
        pass
    def match(self, request):
        path = request.subpath if hasattr(request, 'subpath') else request.path
        if str(self.path) != '' and str(self.path) != '/':
            if path[:len(self.path)] != self.path:
                return False
            else:
                path = path[len(self.path):]
                request.subpath = path
        return self.router.match(request)

def _re_flatten(p):
    if '(' not in p:
        return p
    return re.sub(
        r'(\\*)(\(\?P<[^>]+>|\((?!\?))',
        lambda m: m.group(0) if len(m.group(1)) % 2 else m.group(1) + '(?:',
        p
    )

class Path():
    rule_syntax = re.compile('(\\\\*)(?:(?:<([a-zA-Z_][a-zA-Z_0-9]*)?(?::([a-zA-Z_]*)(?::((?:\\\\.|[^\\\\>])+)?)?)?>))')
    filters = {
        're': lambda conf: (_re_flatten(conf or '[^/]+'), None, None),
        'int': lambda conf: (r'-?\d+', int, lambda x: str(int(x))),
        'float': lambda conf: (r'-?[\d.]+', float, lambda x: str(float(x))),
        'path': lambda conf: (r'.+?', None, None)
    }
    default_filter = 're'
    def __init__(self, path):
        self.raw = path
        self.builder = []
        anons = 0
        pattern = ''
        keys = []
        filters = []
        is_static = True
        for key, mode, conf in self._itertoken(path):
            if mode:
                is_static = False
                if mode == 'default':
                    mode = self.default_filter
                mask, in_filter, out_filter = self.filters[mode](conf)
                if not key:
                    pattern+='(?:%s)' % mask
                    key = 'anon%d' % anons
                    anons += 1
                else:
                    pattern += '(?P<%s>%s)' % (key, mask)
                    keys.append(key)
                if in_filter:
                    filters.append((key, in_filter))
            elif key:
                pattern += re.escape(key)

        try:
            re_pattern = re.compile('^(%s)$' % pattern)
            re_match = re_pattern.match
        except re.error as e:
            raise RouteSyntaxError('')

        if filters:
            def getargs(path):
                url_args = re_match(path).groupdict()
                for name, wildcard_filter in filters:
                    try:
                        url_args[name] = wildcard_filter(url_args[name])
                    except ValueError:
                        raise RoutingError()
                return url_args
        elif re_pattern.groupindex:
            def getargs(path):
                return re_match(path).groupdict()
        else:
            getargs = None
        flatpat = _re_flatten(pattern)
        self.rule = (path, re.compile('(^%s$)' % flatpat), getargs)
    def _itertoken(self, path):
        offset = 0
        prefix = ''
        for match in self.rule_syntax.finditer(path):
            prefix += path[offset:match.start()]
            g = match.groups()
            if len(g[0]) % 2:
                prefix += match.group(0)[len(g[0]):]
                offset = match.end()
                continue
            if prefix:
                yield prefix, None, None
            name, filtr, conf = g[1:4]
            yield name, filtr or 'default', conf or None
            offset, prefix = match.end(), ''
        if offset<=len(path) or prefix:
            yield prefix + path[offset:], None, None
    def match(self, target):
        path, flatpat, getargs = self.rule
        if flatpat.match(target) != None:
            return getargs(target) if getargs else {}
        else:
            return False



from mitama._extra import _Singleton

class HookRegistry(_Singleton):
    def __init__(self):
        self.create_user_hooks = list()
        self.create_group_hooks = list()
        self.update_user_hooks = list()
        self.update_group_hooks = list()
        self.delete_user_hooks = list()
        self.delete_group_hooks = list()
    def create_user(self, target):
        for func in self.create_user_hooks:
            func(target)
    def create_group(self, target):
        for func in self.create_group_hooks:
            func(target)
    def update_user(self, target):
        for func in self.update_user_hooks:
            func(target)
    def update_group(self, target):
        for func in self.update_group_hooks:
            func(target)
    def delete_user(self, target):
        for func in self.delete_user_hooks:
            func(target)
    def delete_group(self, target):
        for func in self.delete_group_hooks:
            func(target)
    def add_create_user_hook(self, func):
        self.create_user_hooks.append(func)
    def add_create_group_hook(self, func):
        self.create_group_hooks.append(func)
    def add_update_user_hook(self, func):
        self.update_user_hooks.append(func)
    def add_update_group_hook(self, func):
        self.update_group_hooks.append(func)
    def add_delete_user_hook(self, func):
        self.delete_user_hooks.append(func)
    def add_delete_group_hook(self, func):
        self.delete_group_hooks.append(func)



from mitama.http import Response
from mitama.nodes import User
from mitama.app import Middleware

class InitializeMiddleware(Middleware):
    def process(self, request, handler):
        if User.query.count() == 0:
            return Response.redirect(self.app.convert_url('/setup'))
        else:
            return handler(request)




#!/usr/bin/python
'''プロジェクト作成コマンド

    * ディレクトリ作ってmitama.json置くだけのコマンド

'''

import os
from pathlib import Path
from .init import init_project_dir

class Command:
    def handle(self, argv = None):
        try:
            project_name = argv[0]
        except IndexError:
            raise IndexError(
                'No project name given to command arguments.'
            )
        current_dir = Path(os.getcwd())
        project_dir = current_dir / project_name
        os.mkdir(project_dir)
        init_project_dir(project_dir)



#!/usr/bin/python
'''ベースモデル

    * アプリサイドから触るときに継承させるベースクラス
    * Djangoのdjango.db.modelsみたいなイメージ
    * データベースはユーザー管理をしているものとは別にしたいので、そのつもりで
    * UserとGroupみたいなカスタム型を使えるようになってると良い
    * Flaskのsqlalchemy拡張が参考に成る
'''

from sqlalchemy.orm import class_mapper
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.types import TypeDecorator
from mitama.db.types import Column, Integer, String, Node, Group, LargeBinary
from mitama._extra import _classproperty
import re

class Model():
    _id = Column(Integer, primary_key = True)
    @_classproperty
    def type(cls):
        class Type(TypeDecorator):
            impl = Integer
            def process_bind_param(self, value, dialect):
                if value == None:
                    return None
                else:
                    return value._id
            def process_result_value(self, value, dialect):
                if value == None:
                    return None
                else:
                    user = cls.retrieve(value)
                    return user
        return Type
    @declared_attr
    def __tablename__(cls):
        return re.sub("(.[A-Z])",lambda x:x.group(1)[0] + "_" +x.group(1)[1], cls.__name__).lower()
    def create(self):
        self.query.session.add(self)
        self.query.session.commit()
    def update(self):
        self.query.session.commit()
    def delete(self):
        self.query.session.delete(self)
        self.query.session.commit()
    @classmethod
    def list(cls, cond = None):
        if cond != None:
            return cls.query.filter(cond).all()
        else:
            return cls.query.filter().all()
    @classmethod
    def retrieve(cls, id = None):
        if id != None:
            node = cls.query.filter(cls._id == id).one()
        else:
            raise Exception('Id not given')
        return node



import time
import json
import base64
from cryptography import fernet
from collections.abc import MutableMapping
from mitama.http import Response

class Session(MutableMapping):
    def __init__(self, identity, *, data, new, max_age = None):
        self._changed = False
        self._mapping = {}
        self._identity = identity if data != {} else None
        self._new = new if data != {} else True
        self._max_age = max_age
        created = data.get('created', None) if data else None
        session_data = data.get('session', None) if data else None
        now = int(time.time())
        age = now - created if created else now
        if max_age is not None and age > max_age:
            session_data = None
        if self._new or created is None:
            self._created = now
        else:
            self._created = created

        if session_data is not None:
            self._mapping.update(session_data)
    @property
    def new(self):
        return self._new
    @property
    def identity(self):
        return self._identity
    @property
    def created(self):
        return self._created
    @property
    def empty(self):
        return not bool(self._mapping)
    @property
    def max_age(self):
        return self._max_age
    @max_age.setter
    def max_age(self, value):
        self._max_age = value
    def changed(self):
        self._changed = True
    def invalidate(self):
        self._changed = True
        self._mapping = {}
    def set_new_identity(self, identity):
        if not self._new:
            raise RuntimeError("Can't change identity for a session which is not new")
        self._identity = identity
    def __len__(self):
        return len(self._mapping)
    def __iter__(self):
        return iter(self._mapping)
    def __containers__(self, key):
        return key in self._mapping
    def __getitem__(self, key):
        return self._mapping[key]
    def __setitem__(self, key, value):
        self._mapping[key] = value
        self._changed = True
    def __delitem__(self, key):
        del self._mapping[key]
        self._changed = True

class EncryptedCookieStorage():
    def __init__(self, secret_key, *, cookie_name = 'MITAMA_SESSION',
                 domain = None, max_age = None, path = '/',
                 secure = None, httponly = True, encoder = json.dumps,
                 decoder = json.loads):
        self._cookie_name = cookie_name
        self._cookie_params = dict(domain = domain,
                                   max_age = max_age,
                                   path = path,
                                   secure = secure,
                                   httponly = httponly)
        self._max_age = max_age
        self._encoder = encoder
        self._decoder = decoder
        if isinstance(secret_key, str):
            pass
        elif isinstance(secret_key, (bytes, bytearray)):
            secret_key = base64.urlsafe_b64encode(secret_key)
        self._fernet = fernet.Fernet(secret_key)
    @property
    def cookie_name(self):
        return self._cookie_name
    @property
    def max_age(self):
        return self._max_age
    @property
    def cookie_params(self):
        return self._cookie_params
    def _get_session_data(self, session):
        if not session.empty:
            data = {
                'created': session.created,
                'session': session._mapping
            }
        else:
            data = {}
        return data
    def load_session(self, request):
        cookie = self.load_cookie(request)
        if cookie is None:
            return Session(None, data = None, new = True, max_age = self.max_age)
        else:
            try:
                data = self._decoder(
                    self._fernet.decrypt(
                        cookie.encode('utf-8'),
                        ttl = self._max_age
                    ).decode('utf-8')
                )
                return Session(None, data = data, new = False, max_age = self.max_age)
            except:
                return Session(None, data = None, new = True, max_age = self.max_age)
    def save_session(self, request, response, session):
        if session.empty:
            return self.save_cookie(response, '', max_age = session.max_age)
        cookie_data = self._encoder(
            self._get_session_data(session)
        ).encode('utf-8')
        self.save_cookie(
            response,
            self._fernet.encrypt(cookie_data).decode('utf-8'),
            max_age = session.max_age)
    def load_cookie(self, request):
        cookie = request.cookies.get(self._cookie_name)
        return cookie
    def save_cookie(self, response, cookie_data, *, max_age = None):
        params = dict(self._cookie_params)
        params_ = dict()
        for k in params.keys():
            params_[k.replace('_', '-')] = params[k]
        if max_age is not None:
            params_['max-age'] = max_age
            params_['expires'] = time.strftime('%a, %d-%b-%Y %T GMT', time.gmtime(time.time() + max_age))
        if not cookie_data:
            response.del_cookie(
                self._cookie_name,
                domain = params_['domain'],
                path = params_['path']
            )
        else:
            response.set_cookie(self._cookie_name, cookie_data, **params_)





from mitama.http import get_session, Response
from .model import *

async def home(request):
    print('home')
    sess = await get_session(request)
    if 'count' not in sess:
        sess['count'] = 0
    else:
        sess['count'] += 1
    return Response(text = 'new some object '+str(sess['count']))

async def about(request):
    return Response(text = 'this is about')



import base64, re
import ssl
#from http.server import ThreadingHTTPServer
from http.server import HTTPServer
from socketserver import StreamRequestHandler
from .request import Request
from .response import Response
import threading


def run_app(app, port, request_factory = Request.parse_stream, ssl = False):
    class RequestHandler(StreamRequestHandler):
        def handle(self):
            request = request_factory(self.rfile)
            response = app(request)
            response.start(request, self.wfile)
    class SSLRequestHandler(StreamRequestHandler):
        def handle(self):
            request = request_factory(self.rfile, True)
            response = app(request)
            response.start(request, self.wfile)
    def serve():
        #with ThreadingHTTPServer(('', int(port)), RequestHandler) as server:
        with HTTPServer(('', int(port)), RequestHandler) as server:
            server.serve_forever()
    th_nossl = threading.Thread(name = 'nossl', target = serve)
    if ssl != False:
        def ssl_serve():
            #with ThreadingHTTPServer(('', int(ssl['port'])), SSLRequestHandler) as server:
            with HTTPServer(('', int(ssl['port'])), SSLRequestHandler) as server:
                if ssl!=False:
                    ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                    crtx.load_cert_chain(ssl['cert'], ssl['key'])
                    ctx.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
                    server.socket = ctx.wrap_socket(server.socket)
                server.serve_forever()
        th_ssl = threading.Thread(name = 'ssl', target = ssl_serve)
        th_ssl.start()
        th_nossl.start()
    else:
        th_nossl.start()



from mitama.db import _Database
from mitama.db.driver.sqlite3 import get_test_engine
import bcrypt

def test_auth():
    db = _Database()
    db.set_engine(get_test_engine())
    from mitama.nodes import User
    from mitama.auth import password_hash, password_auth
    db.create_all()
    user = User()
    user._id = 123
    user.name = 'someone'
    user.screen_name = 'somebody'
    user.password = password_hash('somephrase')
    db.session.add(user)
    db.session.commit()
    assert password_auth('somebody', 'somephrase')
    try:
        password_auth('somebody', 'anyphrase')
        result = True
    except:
        result = False
    assert not result

def test_jwt():
    from mitama.nodes import User
    from mitama.auth import get_jwt, check_jwt
    user = User.query.first()
    assert check_jwt(get_jwt(user)) == user



#!/usr/bin/python
'''configの実装

柔軟性を上げたかったので、デフォルト値を持ったオブジェクトのプロパティをdictの値で更新する方式をとった
とりあえず、プロジェクトフォルダ直下のmitama.jsonを読む仕様にしている。
'''

import os
from pathlib import Path
import json

class Config:
    apps = dict()
    def __init__(self, path, dic):
        self._project_dir = path;
        self._sqlite_db_path = Path(self._project_dir) / 'db.sqlite3'
        for k in dic:
            try:
                setattr(self, k, dic[k])
            except:
                continue
    def to_dict(self):
        # dictに変換する
        dic = dict()
        for k in self.__dict__:
            if k[0] != '_':  #__dict__では_Configプレフィクスがつくので、その文字数を避けてる
                dic[k] = self.__dict__[k]
        return dic


def get_from_project_dir():
    '''プロジェクトフォルダを返します

    App起動時にBuilderからメタ情報が登録されるので、基本的にはアプリから触る必要はないはずです。
    :return: Configインスタンス
    '''
    path = Path(os.getcwd())
    with open(path / 'mitama.json') as f:
        data = f.read()
    return Config(path, json.loads(data))



#!/usr/bin/python
'''サーバー起動コマンド

ポート番号を引数に取ってHTTPサーバーを起動するコマンド
実行されてないマイグレーションもこいつが実行する
'''
from mitama.http import run_app
import mitama.nodes
from mitama.app import _MainApp, AppRegistry
from mitama.conf import get_from_project_dir

class Command:
    def handle(self, argv = None):
        config = get_from_project_dir()
        try:
            port = argv[0]
        except IndexError:
            if hasattr(config, 'port'):
                port = config.port
            else:
                port = '8080'
        if not hasattr(config, 'ssl'):
            config.ssl = False
        app_registry = AppRegistry()
        app_registry.load_config()
        app = _MainApp(app_registry)
        run_app(app, port)



from mitama.conf import Config
import os

def test_config():
    testconf = {
        'apps': [
            {
                'include': 'app1',
                'path': '/'
            },
            {
                'include': 'app2',
                'path': '/app2'
            }
        ]
    }
    conf = Config(os.getcwd(), testconf)
    assert testconf == conf.to_dict()



#!/usr/bin/python
'''コマンド定義

    * 1ファイル1コマンドでクラスとして書いてます
    * handleがコマンドライン引数の引数部を受け取って実行する
    * 抽象化はするかどうか迷ってる（Pythonで抽象化がどれくらい尊いものかよくわかんない）
    * argparseを入れようか迷ったけど、位置引数しか必要に成る予定が無いのでとりあえず入れてない
'''

import importlib
import sys

def exec():
    '''コマンドを起動します

    コマンドライン引数からサブコマンドのインスタンスを生成し、起動します。
    '''
    subcmd = sys.argv[1]
    if subcmd == '':
        subcmd = 'help'
    m = importlib.import_module('.' + subcmd, 'mitama.command')
    cmd = m.Command()
    cmd.handle(sys.argv[2:])




#!/usr/bin/python
'''ノード定義

    * UserとGroupのモデル定義を書きます。
    * 関係テーブルのモデル実装は別モジュールにしようかと思ってる
    * sqlalchemyのベースクラスを拡張したNodeクラスに共通のプロパティを載せて、そいつらをUserとGroupに継承させてます。

Todo:
    * sqlalchemy用にUser型とGroup型を作って、↓のクラスをそのまま使ってDB呼び出しできるようにしたい
'''

from sqlalchemy.ext.declarative import declarative_base
from mitama.db import _CoreDatabase, func, orm
from mitama.db.types import Column, Integer, String, Node, Group, LargeBinary
from mitama.hook import HookRegistry
from mitama.noimage import load_noimage_user, load_noimage_group

db = _CoreDatabase()
hook_registry = HookRegistry()

class Relation(db.Model):
    parent = Column(Group)
    child = Column(Node)

class Node(object):
    _icon = Column(LargeBinary)
    name = Column(String(255))
    screen_name = Column(String(255))
    @property
    def id(self):
        return self._id
    @classmethod
    def retrieve(cls, id = None, screen_name = None):
        if id != None:
            node = cls.query.filter(cls._id == id).first()
        elif screen_name != None:
            node = cls.query.filter(cls.screen_name == screen_name).first()
        else:
            raise Exception('')
        return node
    def icon_to_dataurl(self):
        f = magic.Magic(mime = True, uncompress = True)
        mime = f.from_buffer(self._icon)
        return 'data:'+mime+';base64,'+b64encode(self.icon).decode()
    def parents(self):
        rels = Relation.query.filter(Relation.child == self).all()
        parent = list()
        for rel in rels:
            parent.append(rel.parent)
        return parent
    def is_ancestor(self, node):
        if not isinstance(node, Group) and not isinstance(node, User):
            raise TypeError('Checking object must be Group or User instance')
        layer = self.parents()
        while len(layer) > 0:
            if isinstance(node, Group) and node in layer:
                return True
            else:
                for node_ in layer:
                    if isinstance(node, User) and node_.is_in(node):
                        return True
            layer_ = list()
            for node_ in layer:
                layer_.extend(
                    node_.parents()
                )
            layer = layer_
        return False

class User(Node, db.Model):
    '''ユーザーのモデルクラスです

    :param _id: 固有のID
    :param screen_name: ログイン名
    :param name: 名前
    :param password: パスワード
    :param icon: アイコン
    '''
    __tablename__ = 'mitama_user'
    password = Column(String(255))
    @property
    def icon(self):
        if self._icon != None:
            return self._icon
        else:
            return load_noimage_user()
    @icon.setter
    def icon(self, value):
        self._icon = value
    def delete(self):
        '''ユーザーを削除します'''
        hook_registry.delete_user(self)
        super().delete()
    def update(self):
        '''ユーザー情報を更新します'''
        super().update()
        hook_registry.update_user(self)
    def create(self):
        '''ユーザーを作成します'''
        super().create()
        hook_registry.create_user(self)

class Group(Node, db.Model):
    '''グループのモデルクラスです

    :param _id: 固有のID
    :param screen_name: ドメイン名
    :param name: 名前
    :param icon: アイコン
    '''
    __tablename__ = 'mitama_group'
    @property
    def icon(self):
        if self._icon != None:
            return self._icon
        else:
            return load_noimage_group()
    @icon.setter
    def icon(self, value):
        self._icon = value
    @classmethod
    def tree(cls):
        noparent = [rel.child._id for rel in db.session.query(Relation.child).group_by(Relation.child) if isinstance(rel.child, Group)]
        groups = [group for group in Group.query.filter().all() if group._id not in noparent and isinstance(group, Group)]
        return groups
    def append(self, node):
        if not isinstance(node, Group) and not isinstance(node, User):
            raise TypeError('Appending object must be Group or User instance')
        rel = Relation()
        rel.parent = self
        rel.child = node
        rel.create()
    def append_all(self, nodes):
        for node in nodes:
            if not isinstance(node, Group) and not isinstance(node, User):
                raise TypeError('Appending object must be Group or User instance')
            rel = Relation()
            rel.parent = self
            rel.child = node
            Relation.query.session.add(rel)
        Relation.query.session.commit()
    def remove(self, node):
        if not isinstance(node, Group) and not isinstance(node, User):
            raise TypeError('Removing object must be Group or User instance')
        rel = Relation.query.filter(Relation.parent == self).filter(Relation.child == node).first()
        rel.delete()
    def remove_all(self, nodes):
        for node in nodes:
            if not isinstance(node, Group) and not isinstance(node, User):
                raise TypeError('Appending object must be Group or User instance')
        rels = Relation.query.filter(Relation.parent == self).filter(Relation.child in nodes).all()
        Relation.query.session.delete(rels)
        Relation.query.session.commit()
    def children(self):
        rels = Relation.query.filter(Relation.parent == self).all()
        children = list()
        for rel in rels:
            children.append(rel.child)
        return children
    def is_descendant(self, node):
        if node.__class__.__name__ != 'Group' and node.__class__.__name__ != 'User':
            raise TypeError('Checking object must be Group or User instance')
        layer = self.children()
        while len(layer) > 0:
            if node in layer:
                return True
            layer_ = list()
            for node_ in layer:
                if node_.__class__.__name__ == 'Group':
                    layer_.extend(
                        node_.children()
                    )
            layer = layer_
        return False
    def is_in(self, node):
        if not isinstance(node, Group) and not isinstance(node, User):
            raise TypeError('Checking object must be Group or User instance')
        rels = Relation.query.filter(Relation.parent == self).filter(Relation.child == node).all()
        return len(rels) != 0
    def delete(self):
        '''グループを削除します'''
        hook_registry.delete_group(self)
        super().delete()
    def update(self):
        '''グループの情報を更新します'''
        super().update()
        hook_registry.update_group(self)
    def create(self):
        '''グループを作成します'''
        super().create()
        hook_registry.create_group(self)



db.create_all()



#!/usr/bin/python
'''パスワード認証

Mitamaではパスワードのハッシュ化にはBCRYPTアルゴリズムの$2yプリフィクスのものを利用します
理由は、前回バージョンの第三版がPHPのpassword_hash関数を使用しており、移植性を保ちたかったからです。
しょうもなくてすみません
'''

import bcrypt
import jwt
import random
import secrets
import hashlib
import base64
from mitama.nodes import User

secret = secrets.token_hex(32)

class AuthorizationError(Exception):
    pass

def password_auth(screen_name, password):
    '''ログイン名とパスワードで認証します

    :param screen_name: ログイン名
    :param password: パスワード
    :return: Userインスタンス
    '''
    user = User.query.filter(User.screen_name == screen_name).first()
    password = base64.b64encode(
        hashlib.sha256(
            password.encode() * 10
        ).digest()
    )
    if bcrypt.checkpw(password, user.password):
        return user
    else:
        raise AuthorizationError('Wrong password')

def password_hash(password):
    '''パスワードをハッシュ化します

    :param password: パスワードのプレーンテキスト
    :return: パスワードハッシュ
    '''
    salt = bcrypt.gensalt()
    password = base64.b64encode(
        hashlib.sha256(
            password.encode() * 10
        ).digest()
    )
    return bcrypt.hashpw(password, salt)

def get_jwt(user):
    '''UserインスタンスからJWTを生成します

    :param user: Userインスタンス
    :return: JWT
    '''
    nonce = ''.join([str(random.randint(0,9)) for i in range(16)])
    result = jwt.encode(
        {
            'id': user._id,
            'nonce': nonce
        },
        secret,
        algorithm='HS256'
    )
    return result.decode()

def check_jwt(token):
    '''JWTからUserインスタンスを取得します

    :param token: JWT
    :return: Userインスタンス
    '''
    try:
        result = jwt.decode(
            token,
            secret,
            algorithm='HS256'
        )
    except jwt.exceptions.InvalidTokenError as err:
        raise AuthorizationError('Invalid token.')
    return User.query.filter(User._id == result['id']).first()



#!/usr/bin/python
'''sqliteのエンジン

    * プロジェクトのDBはプロジェクトのconfigから場所を取って参照する。
    * アプリのDBはアプリのフォルダ直下を参照する
'''

from sqlalchemy import create_engine

def get_engine():
    from mitama.conf import get_from_project_dir
    config = get_from_project_dir()
    return create_engine('sqlite:///'+str(config._sqlite_db_path))

def get_test_engine():
    return create_engine('sqlite://')

def get_app_engine(appname):
    from mitama.conf import get_from_project_dir
    config = get_from_project_dir()
    return create_engine('sqlite:///'+str(config._project_dir / (appname + '/db.sqlite3')))



from http.server import BaseHTTPRequestHandler
import http.cookies
from abc import ABCMeta, abstractmethod

class ResponseBase(metaclass = ABCMeta):
    _version = 'HTTP/1.1'
    def __init__(self, status = 200, reason = None, headers = {}):
        self.headers = headers
        self._status = int(status)
        if reason is None:
            try:
                self._reason = BaseHTTPRequestHandler.responses[self._status][0]
            except:
                self._reason = ''
        else:
            self._reason = reason
        self._cookies = http.cookies.SimpleCookie()
    def set_cookie(self, key, value = None, **kwargs):
        self._cookies[key] = value or ''
        for k,v in kwargs.items():
            self._cookies[key][k] = v or ''
    @abstractmethod
    def start(self, request, stream):
        pass

class Response(ResponseBase):
    def __init__(self, body = None, text = None, headers = {}, status = 200, reason = None, content_type = None, charset = None):
        super().__init__(headers = headers, status = status, reason = reason)
        if body is not None and text is not None:
            raise ValueError('body and text are not allowed together')
        if text is not None:
            if not isinstance(text, str):
                raise TypeError('')
            if content_type is None:
                content_type = 'text/html'
            if charset is None:
                charset = 'utf-8'
            body = text.encode(charset)
            text = None
        else:
            if content_type is not None:
                if charset is not None:
                    content_type += '; charset='+charset
        self.content_type = content_type
        self.body = body
    def start(self, request, stream):
        stream.write(('%s %s %s\r\n' % (self._version, self._status, self._reason)).encode())
        for k,v in self.headers.items():
            stream.write(('%s: %s\r\n' % (k, v)).encode())
        cookies = self._cookies.output().encode()
        if len(cookies) > 0:
            stream.write(cookies)
            stream.write(b'\r\n')
        stream.write(('Content-Type: %s\r\n' % self.content_type).encode())
        stream.write('\r\n'.encode())
        if callable(self.body):
            stream.write(self.body(request))
        elif self.body is not None:
            stream.write(self.body)
    @classmethod
    def render(cls, template, values = {}, **kwargs):
        '''HTMLを描画するレスポンスを返却します

        Jinja2のテンプレートにデータを入れてHTMLを生成し、Responseインスタンスを作成して返却します。
        テンプレート内部からはRequest内の非同期関数を呼び出すことができます。
        :param template: Jinja2テンプレート
        :param request: Requestインスタンス
        :param values: プレースホルダに入力するデータの辞書
        :param **kwargs: aiohttp.web.Responseに受け渡す、statusやheadersなどのデータ
        '''
        def render(request):
            values['request'] = request
            return template.render(values).encode()
        return cls(body = render, **kwargs)
    @classmethod
    def redirect(cls, uri, status=302):
        '''リダイレクトします

        :param uri: リダイレクト先のURL
        :param status: リダイレクト時のステータスコード
        '''
        return cls(headers = {
            'Location': uri
        }, status = status)

class StreamResponse(ResponseBase):
    def __init__(self, body = None, headers = {}, status = 200, reason = None, content_type = None, charset = None):
        super().__init__(headers = headers, status = status, reason = reason)
        self.content_type = content_type
        if charset is not None:
            self.content_type += '; charset='+charset
        if body is not None:
            self.body = body
        else:
            self.body = list()
    def start(self, request, stream):
        stream.write(('%s %s %s\r\n' % (self._version, self._status, self._reason)).encode())
        for k,v in self.headers.items():
            stream.write(('%s: %s\r\n' % (k, v)).encode())
        stream.write(self._cookies.output().encode())
        stream.write(('Content-Type: %s\r\n' % self.content_type).encode())
        stream.write('\r\n'.encode())
        for chunk in self.body:
            if callable(chunk):
                stream.write(chunk())
            else:
                stream.write(chunk)




from pathlib import Path
import magic, os
from mitama.conf import get_from_project_dir
from base64 import b64encode

def load_noimage_app():
    '''アプリのNoImage画像を取得します'''
    config = get_from_project_dir()
    if (config._project_dir / "noimage_app").is_file():
        noimage_app_path = config._project_dir / "static/noimage_app.png"
    else:
        noimage_app_path = Path(os.path.dirname(__file__)) / "app/static/noimage_app.png"
    with open(noimage_app_path, "rb") as f:
        noimage_app = f.read()
    return noimage_app

def load_noimage_user():
    '''ユーザーのNoImage画像を取得します'''
    config = get_from_project_dir()
    if (config._project_dir / "noimage_user").is_file():
        noimage_user_path = (config._project_dir / "static/noimage_user.png")
    else:
        noimage_user_path = (Path(os.path.dirname(__file__)) / "app/static/noimage_user.png")

    with open(noimage_user_path, "rb") as f:
        noimage_user = f.read()
    return noimage_user

def load_noimage_group():
    '''グループのNoImage画像を取得します'''
    config = get_from_project_dir()
    if (config._project_dir / "noimage_group").is_file():
        noimage_group_path = (config._project_dir / "static/noimage_group.png")
    else:
        noimage_group_path = (Path(os.path.dirname(__file__)) / "app/static/noimage_group.png")

    with open(noimage_group_path, "rb") as f:
        noimage_group = f.read()
    return noimage_group




import cgi
import http
import http.cookies
import io
from urllib.parse import parse_qs
from yarl import URL

class _Cookies():
    def __init__(self, cookies):
        self._cookies = cookies
    def __getitem__(self, key):
        if key in self._cookies:
            return self._cookies[key].value
        else:
            raise KeyError('')
    def get(self, key):
        if key in self._cookies:
            return self._cookies[key].value
        else:
            return None

class _RequestPayload():
    def __init__(self, field_storage):
        self._field_storage = field_storage
    def __contains__(self, key):
        if key not in self._field_storage:
            return False
        elif isinstance(self._field_storage[key].file, io.BytesIO):
            return (self._field_storage[key].filename or '') != ''
        else:
            return len(self._field_storage[key].value) > 0
    def __getitem__(self, key):
        if key not in self._field_storage:
            raise KeyError
        elif (self._field_storage[key].filename or '') != '':
            return self._field_storage[key]
        elif len(self._field_storage[key].value) > 0:
            return self._field_storage[key].value
        else:
            raise KeyError
    def get(self, key, default = None):
        if key in self:
            return self[key]
        else:
            return default
    def getlist(self, key, default = list()):
        if key in self:
            return self._field_storage.getlist(key)
        else:
            return default
    @classmethod
    def parse_body(cls, rfile, content_type, length):
        environ = {
            'REQUEST_METHOD': 'POST'
        }
        headers = {
            'content-type': content_type,
            'content-length': length
        }
        fs = cgi.FieldStorage(rfile, environ = environ, headers = headers)
        payload = cls(fs)
        return payload

class Request():
    MessageClass = http.client.HTTPMessage
    def __init__(self, method, path, version, headers, ssl, rfile):
        self._method = method
        self._raw_path = path
        self._url = URL(path)
        path = path.split('?')
        if len(path) == 2:
            path, query = path
            self._path = path
            self._query = dict()
            query = parse_qs(query)
            for k, v in query.items():
                if len(v) == 1:
                    self._query[k] = v[0]
                else:
                    self._query[k] = v
        else:
            self._path = path[0]
            self._query = {}
        self._version = version
        self._headers = headers
        self._secure = ssl
        self._host = self._headers.get('Host')
        self._rfile = rfile
        self._data = dict()
    def __setitem__(self, key, value):
        self._data[key] = value
    def __getitem__(self, key):
        return self._data[key]
    def __delitem__(self, key):
        del self._data[key]
    def get(self, key):
        return self._data[key] if key in self._data else None
    @property
    def scheme(self):
        return 'https' if self._secure else 'http'
    @property
    def host(self):
        return self._host
    @property
    def method(self):
        return self._method
    @property
    def raw_path(self):
        return self._raw_path
    @property
    def path(self):
        return self._path
    @property
    def query(self):
        return self._query
    @property
    def version(self):
        return self._version
    @property
    def headers(self):
        return self._headers
    @property
    def url(self):
        return self._raw_path
    @property
    def cookies(self):
        if hasattr(self, '_cookies'):
            return self._cookies
        else:
            C = http.cookies.BaseCookie()
            cookie_headers = self._headers.get_all('cookie', [])
            cookie_str = '; '.join(cookie_headers)
            C.load(cookie_str)
            self._cookies = _Cookies(C)
            return self._cookies
    def post(self):
        content_type = self._headers.get('Content-Type', '')
        length = int(self._headers.get('Content-Length', 0))
        if content_type.startswith('multipart/form-data') or content_type.startswith('application/x-www-form-urlencoded'):
            parsed_body = _RequestPayload.parse_body(self._rfile, content_type, length)
        elif content_type == 'application/json':
            payload = self._rfile.read()
            parsed_body = json.loads(payload)
        else:
            parsed_body = None
        return parsed_body
    def session(self):
        '''セッション情報を取得します

        :return: セッションの辞書データ
        '''
        sess = self.get('mitama_session')
        if sess is None:
            storage = self.get('mitama_session_storage')
            sess = storage.load_session(self)
            self['mitama_session'] = sess
        return sess
    @classmethod
    def parse_stream(cls, rfile, ssl = False):
        words = rfile.readline().decode().rstrip('\r\n').split(' ')
        version = words[-1]
        if not 2 <= len(words) <= 3:
            raise Exception('Bad Request')
        command, path = words[:2]
        if len(words) == 2:
            close_connection = True
            if command!='GET':
                raise Exception('Bad Request')
        try:
            headers = http.client.parse_headers(rfile, _class = cls.MessageClass)
        except http.client.LineTooLong as err:
            raise Exception('')
        except http.client.HTTPException as err:
            raise Exception('')
        return cls(command, path, version, headers, ssl, rfile)



#!/usr/bin/python
'''データベース

データベースの接続とか抽象化の処理を書きます
Databaseはシングルトンの接続のインスタンスを生成するクラスです
各アプリにはDatabaseを継承したクラスを定義してもらい、そいつのModelプロパティのベースクラスからモデルを作ってもらいます。
'''

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative.api import DeclarativeMeta
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.orm.exc import UnmappedClassError
from sqlalchemy.sql import func
from sqlalchemy import orm
from .model import Model
from .driver.sqlite3 import get_engine, get_app_engine
from mitama._extra import _Singleton
import inspect

class _QueryProperty:
    def __init__(self, db):
        self.db = db
    def __get__(self, obj, type):
        try:
            mapper = orm.class_mapper(type)
            if mapper:
                return type.query_class(
                    mapper,
                    session = self.db.session()
                )
        except UnmappedClassError:
            return None

class _Database(_Singleton):
    engine = None
    session = None
    def __init__(self, model = None, metadata = None, query_class = orm.Query):
        self.Query = query_class
        self.Model = self.make_declarative_base(model, metadata)
    def set_engine(self, engine):
        self.engine = engine
        self.session = scoped_session(
            sessionmaker(
                autocommit = False,
                autoflush = False,
                bind = engine
            )
        )
    def make_declarative_base(self, model = None, metadata = None):
        if model == None:
            model = Model
        if not isinstance(model, DeclarativeMeta):
            model = declarative_base(
                cls = model,
                name = 'Model',
                metadata = metadata
            )
        if metadata is not None and model.metadata is not metadata:
            model.metadata = metadata
        if not getattr(model, 'query_class', None):
            model.query_class = self.Query
        model.query = _QueryProperty(self)
        return model
    def session(self):
        return self.session
    def create_all(self):
        self.Model.metadata.create_all(self.engine)

class _CoreDatabase(_Database):
    def __init__(self, engine = None):
        super().__init__()
        if self.engine == None:
            if engine == None:
                self.set_engine(get_engine())
            else:
                self.set_engine(engine)

class BaseDatabase(_Database):
    '''アプリで利用するデータベースの操作を行うクラス

    アプリからデータベースを使うたい場合、このクラスを継承したクラスをアプリ内に定義します。
    '''
    def __init__(self, engine = None):
        super().__init__()
        if self.engine == None:
            if engine == None:
                package_name = inspect.getmodule(self.__class__).__package__
                self.set_engine(get_app_engine(package_name))
            else:
                self.set_engine(engine)



#!/usr/bin/python
'''アプリ作成コマンド

    * アプリのテンプレートを作成する

'''

import os
from pathlib import Path
import shutil
from .init import init_project_dir

class Command:
    def handle(self, argv = None):
        try:
            project_name = argv[0]
        except IndexError:
            raise IndexError(
                'No app name given to command arguments.'
            )
        current_dir = Path(os.getcwd())
        project_dir = current_dir / project_name
        src = Path(os.path.dirname(__file__)) / '../skeleton'
        shutil.copytree(src, project_dir, symlinks=False)



#!/usr/bin/python

class Command:
    def handle(self, argv = None):
        print('usage: mitama <command> [<args>]')
        print('')
        print('new [name]       Create new Mitama project')
        print('init             Initialize directory as Mitama project')
        print('createapp [name] Create new Mitama application package')
        print('run [port]       Launch server in specified port')
        print('version          Show version of Mitama')



#!/usr/bin/python
'''型定義

    * ノリで書いたから自分でもよくわからないけど、データベースのカスタム型を定義してみた
    * イメージとしては、データベースの実体には各モデルのidプロパティが入る
    * 取り出すときは数字からそれに対応するモデルのインスタンスが還ってくる
    * 当然、↓のコードはUser.getなんて関数は用意してないので動かないと思われ

'''

from sqlalchemy.types import TypeDecorator
from sqlalchemy import *

class User(TypeDecorator):
    impl = Integer
    def process_bind_param(self, value, dialect):
        if value == None:
            return None
        else:
            return value._id
    def process_result_value(self, value, dialect):
        from mitama.nodes import User
        if value == None:
            return None
        else:
            user = User.retrieve(value)
            return user

class Group(TypeDecorator):
    impl = Integer
    def process_bind_param(self, value, dialect):
        if value == None:
            return None
        else:
            return value._id
    def process_result_value(self, value, dialect):
        from mitama.nodes import Group
        if value == None:
            return None
        else:
            group = Group.retrieve(value)
            return group

class Node(TypeDecorator):
    impl = Integer
    def process_bind_param(self, value, dialect):
        if value.__class__.__name__ == 'Group':
            return value._id * 2
        elif value.__class__.__name__ == 'User':
            return value._id * 2 - 1
        else:
            raise TypeError('Appending object must be Group or User instance')
    def process_result_value(self, value, dialect):
        if value % 2 == 1:
            from mitama.nodes import User
            node = User.retrieve((value + 1) / 2)
        else:
            from mitama.nodes import Group
            node = Group.retrieve(value / 2)
        return node



from mitama.app import Controller, AppRegistry
from mitama.http import Response
from mitama.nodes import User, Group
from mitama.auth import password_hash, password_auth, get_jwt, AuthorizationError
from mitama.noimage import load_noimage_group, load_noimage_user
import json
import traceback
from uuid import uuid4
from .model import Invite, CreateUserPermission, UpdateUserPermission, DeleteUserPermission, CreateGroupPermission, UpdateGroupPermission, DeleteGroupPermission, Admin

class SessionController(Controller):
    def login(self, request):
        template = self.view.get_template('login.html')
        if request.method == 'POST':
            try:
                post = request.post()
                result = password_auth(post['screen_name'], post['password'])
                sess = request.session()
                sess['jwt_token'] = get_jwt(result)
                redirect_to = request.query.get('redirect_to', '/')
                return Response.redirect(
                    redirect_to
                )
            except AuthorizationError as err:
                error = 'パスワード、またはログイン名が間違っています'
                return Response.render(
                    template,
                    {
                        'error':error
                    },
                    status = 401
                )
        return Response.render(
            template,
            status = 401
        )

    def logout(self, request):
        sess = request.session()
        sess['jwt_token'] = None
        redirect_to = request.query.get('redirect_to', '/')
        return Response.redirect(redirect_to)

class RegisterController(Controller):
    def signup(self, request):
        sess = request.session()
        template = self.view.get_template('signup.html')
        invite = Invite.query.filter(Invite.token == request.query["token"]).first()
        if request.method == "POST":
            try:
                data = request.post()
                user = User()
                user.password = password_hash(data['password'])
                if invite.editable:
                    user.screen_name = data.get('screen_name', '')
                    user.name = data.get('name', '')
                    user.icon = data['icon'].file.read() if "icon" in data else invite.icon
                else:
                    user.screen_name = invite.screen_name
                    user.name = invite.name
                    user.icon = invite.icon
                user.create()
                invite.delete()
                UpdateUserPermission.accept(user, user)
                sess["jwt_token"] = get_jwt(user)
                return Response.redirect(
                    self.app.convert_url('/')
                )
            except Exception as err:
                error = str(err)
                return Response.render(template, {
                    'error': error,
                    "name": data.get("name", invite.name),
                    "screen_name": data.get("screen_name", invite.screen_name),
                    "password": data.get("password", ''),
                    "icon": data.get("icon").file.read(),
                    'editable': invite.editable
                })
        return Response.render(template, {
            "icon": invite.icon,
            "name": invite.name,
            "screen_name": invite.screen_name,
            'editable': invite.editable
        })
    def setup(self, request):
        sess = request.session()
        template = self.app.view.get_template('setup.html')
        if request.method == 'POST':
            try:
                data = request.post()
                user = User()
                user.screen_name = data['screen_name']
                user.name = data['name']
                user.password = password_hash(data['password'])
                user.icon = data["icon"].file.read() if 'icon' in data else load_noimage_user()
                user.create()
                Admin.accept(user)
                CreateUserPermission.accept(user)
                UpdateUserPermission.accept(user)
                DeleteUserPermission.accept(user)
                CreateGroupPermission.accept(user)
                UpdateGroupPermission.accept(user)
                DeleteGroupPermission.accept(user)
                UpdateUserPermission.accept(user, user)
                sess["jwt_token"] = get_jwt(user)
                return Response.redirect(
                    self.app.convert_url("/")
                )
            except Exception as err:
                error = str(err)
                return Response.render(template, {
                    'error': error
                })
        return Response.render(template)
# HomeControllerではユーザー定義のダッシュボード的なのを作れるようにしたいけど、時間的にパス
'''
class HomeController(Controller):
    def handle(self, request):
        template = self.view.get_template('home.html')
        return Response.render(template)
'''

class UsersController(Controller):
    def create(self, req):
        if CreateUserPermission.is_forbidden(req.user):
            return self.app.error(req, 403)
        template = self.view.get_template('user/create.html')
        invites = Invite.list()
        if req.method == 'POST':
            post = req.post()
            try:
                icon = post["icon"].file.read() if "icon" in post else load_noimage_user()
                invite = Invite()
                invite.name = post.get('name', '')
                invite.screen_name = post.get('screen_name', '')
                invite.icon = icon
                invite.token = str(uuid4())
                invite.editable = 'editable' in post
                invite.create()
                invites = Invite.list()
                return Response.render(template, {
                    'invites': invites,
                    "icon": load_noimage_user()
                })
            except Exception as err:
                error = str(err)
                return Response.render(template, {
                    'invites': invites,
                    "name": post.get("name", ''),
                    "screen_name": post.get("screen_name", ''),
                    "icon": icon,
                    'error': error
                })
        return Response.render(template, {
            'invites': invites,
            "icon": load_noimage_user()
        })
    def cancel(self, req):
        if CreateUserPermission.is_forbidden(req.user):
            return self.app.error(req, 403)
        invite = Invite.retrieve(req.params['id'])
        invite.delete()
        return Response.redirect(self.app.convert_url('/users/invite'))
    def retrieve(self, req):
        template = self.view.get_template('user/retrieve.html')
        user = User.retrieve(screen_name = req.params["id"])
        return Response.render(template, {
            "user": user,
            'updatable': UpdateUserPermission.is_accepted(req.user, user)
        })
    def update(self, req):
        template = self.view.get_template('user/update.html')
        user = User.retrieve(screen_name = req.params["id"])
        if UpdateUserPermission.is_forbidden(req.user, user):
            return self.app.error(req, 403)
        if req.method == "POST":
            post = req.post()
            try:
                icon = post["icon"].file.read() if "icon" in post else user.icon
                user.screen_name = post["screen_name"]
                user.name = post["name"]
                user.icon = icon
                user.update()
                if Admin.is_accepted(req.user):
                    if 'user_create' in post:
                        CreateUserPermission.accept(user)
                    else:
                        CreateUserPermission.forbit(user)
                    if 'user_update' in post:
                        UpdateUserPermission.accept(user)
                    else:
                        UpdateUserPermission.forbit(user)
                    if 'user_delete' in post:
                        DeleteUserPermission.accept(user)
                    else:
                        DeleteUserPermission.forbit(user)
                    if 'group_create' in post:
                        CreateGroupPermission.accept(user)
                    else:
                        CreateGroupPermission.forbit(user)
                    if 'group_update' in post:
                        UpdateGroupPermission.accept(user)
                    else:
                        UpdateGroupPermission.forbit(user)
                    if 'group_delete' in post:
                        DeleteGroupPermission.accept(user)
                    else:
                        DeleteGroupPermission.forbit(user)
                    if 'admin' in post:
                        Admin.accept(user)
                    else:
                        Admin.forbit(user)
                return Response.render(template, {
                    "message": "変更を保存しました",
                    "user": user,
                    "screen_name": user.screen_name,
                    "name": user.name,
                    "icon": user.icon,
                })
            except Exception as err:
                error = str(err)
                return Response.render(template, {
                    "error": error,
                    "user": user,
                    "screen_name": post.get("screen_name", user.screen_name),
                    "name": post.get("name", user.name),
                    "icon": icon,
                })
        return Response.render(template, {
            "user": user,
            "screen_name": user.screen_name,
            "name": user.name,
            "icon": user.icon,
        })
    def delete(self, req):
        if DeleteUserPermission.is_forbidden(req.user):
            return self.app.error(req, 403)
        template = self.view.get_template('user/delete.html')
        return Response.render(template)
    def list(self, req):
        template = self.view.get_template('user/list.html')
        users = User.list()
        return Response.render(template, {
            'users': users,
            'create_permission': CreateUserPermission.is_accepted(req.user),
        })

class GroupsController(Controller):
    def create(self, req):
        if CreateGroupPermission.is_forbidden(req.user):
            return self.app.error(request, 403)
        template = self.view.get_template('group/create.html')
        groups = Group.list()
        if req.method == 'POST':
            post = req.post()
            try:
                group = Group()
                group.name = post['name']
                group.screen_name = post['screen_name']
                group.icon = post['icon'].file.read() if "icon" in post else None
                group.create()
                if "parent" in post and post['parent'] != '':
                    Group.retrieve(int(post['parent'])).append(group)
                group.append(req.user)
                UpdateGroupPermission.accept(req.user, group)
                return Response.redirect(self.app.convert_url("/groups"))
            except Exception as err:
                error = str(err)
                return Response.render(template, {
                    'groups': groups,
                    "icon": load_noimage_group(),
                    'error': error
                })
        return Response.render(template, {
            'groups': groups,
            "icon": load_noimage_group()
        })
    def retrieve(self, req):
        template = self.view.get_template('group/retrieve.html')
        group = Group.retrieve(screen_name = req.params["id"])
        return Response.render(template, {
            "group": group,
            'updatable': UpdateGroupPermission.is_accepted(req.user, group)
        })
    def update(self, req):
        template = self.view.get_template('group/update.html')
        group = Group.retrieve(screen_name = req.params["id"])
        groups = list()
        for g in Group.list():
            if not (group.is_ancestor(g) or group.is_descendant(g) or g==group):
                groups.append(g)
        users = list()
        for u in User.list():
            if not group.is_in(u):
                users.append(u)
        if UpdateGroupPermission.is_forbidden(req.user, group):
            return self.app.error(req, 403)
        if req.method == "POST":
            post = req.post()
            try:
                icon = post["icon"].file.read() if "icon" in post else group.icon
                group.screen_name = post["screen_name"]
                group.name = post["name"]
                group.icon = icon
                group.update()
                if Admin.is_accepted(req.user):
                    if 'user_create' in post:
                        CreateUserPermission.accept(group)
                    else:
                        CreateUserPermission.forbit(group)
                    if 'user_update' in post:
                        UpdateUserPermission.accept(group)
                    else:
                        UpdateUserPermission.forbit(group)
                    if 'user_delete' in post:
                        DeleteUserPermission.accept(group)
                    else:
                        DeleteUserPermission.forbit(group)
                    if 'group_create' in post:
                        CreateGroupPermission.accept(group)
                    else:
                        CreateGroupPermission.forbit(group)
                    if 'group_update' in post:
                        UpdateGroupPermission.accept(group)
                    else:
                        UpdateGroupPermission.forbit(group)
                    if 'group_delete' in post:
                        DeleteGroupPermission.accept(group)
                    else:
                        DeleteGroupPermission.forbit(group)
                    if 'admin' in post:
                        Admin.accept(group)
                    else:
                        Admin.forbit(group)
                return Response.render(template, {
                    "message": "変更を保存しました",
                    "group": group,
                    "screen_name": group.screen_name,
                    "name": group.name,
                    'all_groups': groups,
                    'all_users': users,
                    "icon": group.icon,
                })
            except Exception as err:
                error = str(err)
                return Response.render(template, {
                    "error": error,
                    'all_groups': groups,
                    'all_users': users,
                    "group": group,
                    "screen_name": post.get("screen_name", ''),
                    "name": post.get("name", ''),
                    "icon": group.icon,
                })
        return Response.render(template, {
            "group": group,
            'all_groups': groups,
            'all_users': users,
            "screen_name": group.screen_name,
            "name": group.name,
            "icon": group.icon,
        })
    def append(self, req):
        post = req.post()
        try:
            group = Group.retrieve(screen_name = req.params['id'])
            nodes = list()
            if 'user' in post:
                for uid in post.getlist('user'):
                    try:
                        nodes.append(User.retrieve(int(uid)))
                    except Exception as err:
                        print(err)
                        pass
            if 'group' in post:
                for gid in post.getlist('group'):
                    try:
                        nodes.append(Group.retrieve(int(gid)))
                    except Exception as err:
                        print(err)
                        pass
            group.append_all(nodes)
        except Exception as err:
            pass
        finally:
            return Response.redirect(self.app.convert_url('/groups/'+group.screen_name+'/settings'))
    def remove(self, req):
        try:
            group = Group.retrieve(screen_name = req.params['id'])
            cid = int(req.params['cid'])
            if cid % 2 == 0:
                child = Group.retrieve(cid / 2)
            else:
                child = User.retrieve((cid + 1) / 2)
            group.remove(child)
        except Exception as err:
            pass
        finally:
            return Response.redirect(self.app.convert_url('/groups/'+group.screen_name+'/settings'))
    def accept(self, req):
        group = Group.retrieve(screen_name = req.params['id'])
        if UpdateGroupPermission.is_forbidden(req.user, group):
            return self.app.error(req, 403)
        user = User.retrieve(int(req.params['cid']))
        UpdateGroupPermission.accept(user, group)
        return Response.redirect(self.app.convert_url('/groups/'+group.screen_name+'/settings'))
    def forbit(self, req):
        group = Group.retrieve(screen_name = req.params['id'])
        if UpdateGroupPermission.is_forbidden(req.user, group):
            return self.app.error(req, 403)
        user = User.retrieve(int(req.params['cid']))
        UpdateGroupPermission.forbit(user, group)
        return Response.redirect(self.app.convert_url('/groups/'+group.screen_name+'/settings'))
    def delete(self, req):
        if DeleteGroupPermission.is_forbidden(req.user):
            return self.app.error(req, 403)
        template = self.view.get_template('group/delete.html')
        return Response.render(template)
    def list(self, req):
        template = self.view.get_template('group/list.html')
        groups = Group.tree()
        return Response.render(template, {
            'groups': groups,
            'create_permission': CreateGroupPermission.is_accepted(req.user),
        })

class AppsController(Controller):
    def update(self, req):
        if Admin.is_forbidden(req.user):
            return self.app.error(req, 403)
        template = self.view.get_template('apps/update.html')
        apps = AppRegistry()
        if req.method == "POST":
            apps.reset()
            post = req.post()
            try:
                prefix = post["prefix"]
                data = dict()
                data["apps"] = dict()
                for package, path in prefix.items():
                    data["apps"][package] = {
                        "path": path
                    }
                with open(self.app.project_root_dir / "mitama.json", 'w') as f:
                    f.write(json.dumps(data))
                apps.load_config()
                return Response.render(template, {
                    'message': '変更を保存しました',
                    "apps": apps,
                })
            except Exception as err:
                return Response.render(template, {
                    "apps": apps,
                    'error': str(err)
                })
        return Response.render(template, {
            "apps": apps
        })
    def list(self, req):
        template = self.view.get_template('apps/list.html')
        apps = AppRegistry()
        return Response.render(template, {
            "apps": apps,
        })



from mitama.app import App, Router, static_files
from mitama.app.method import view
from .controller import WelcomeController


class App(App):
    #name = 'MyApp'
    #description = 'This is my App.'
    router = Router([
        view('/', WelcomeController),
        view('/static/<path:path>', static_files()),
    ])



#!/usr/bin/python

from setuptools import setup, find_packages

setup(
    name = 'mitama',
    install_requires = ['sqlalchemy', 'bcrypt', 'pyjwt', 'jinja2', 'cryptography', 'python-magic', 'watchdog', 'yarl'],
    extra_requires = {
        'develop': ['pytest']
    },
    use_scm_version = True,
    setup_requires = [
        'setuptools_scm'
    ],
    packages = find_packages(),
    package_data = {
        'mitama.portal': [
            'templates/*.html',
            'templates/**/*.html',
            'static/*',
        ],
        'mitama.skeleton': [
            'templates/*.html',
            'static/*',
        ],
        'mitama.http': [
            'templates/*.html',
        ],
        'mitama.app': [
            'static/*',
        ]
    },
    entry_points = {
        'console_scripts': [
            'mitama = mitama.command:exec'
        ]
    }
)



from mitama.db import _CoreDatabase
from mitama.db.driver.sqlite3 import get_test_engine
from sqlalchemy.orm import *
from mitama.db import types
from mitama.db.types import Column
from pprint import pprint

def create_test_db():
    engine = get_test_engine()
    return engine

def test_nodes():
    engine = create_test_db()
    db = _CoreDatabase(engine)
    from mitama.nodes import User, Group
    class Profile(db.Model):
        __tablename__ = 'test_profile'
        description = Column(types.String(255))
        user = Column(types.User)
    class GroupProfile(db.Model):
        __tablename__ = 'test_group_profile'
        description = Column(types.String(255))
        group = Column(types.Group)
    db.create_all()
    test_user = User()
    test_user.name = 'test'
    test_user.screen_name = 'testuser'
    test_user.password = 'testpass'
    db.session.add(test_user)
    test_group= Group()
    test_group.name = 'test'
    test_group.screen_name = 'testgroup'
    db.session.add(test_group)
    db.session.commit()
    test_user_profile = Profile()
    test_user_profile.description = 'this is the test user'
    test_user_profile.user = test_user
    db.session.add(test_user_profile)
    test_group_profile = GroupProfile()
    test_group_profile.description = 'this is the test group'
    test_group_profile.group = test_group
    db.session.add(test_group_profile)
    db.session.commit()
    test_user_profile_ = Profile.query.first()
    test_group_profile_ = GroupProfile.query.first()
    assert test_user_profile_.user.name == test_user.name
    assert test_user_profile_.user.screen_name == test_user.screen_name
    assert test_user_profile_.user.password == test_user.password
    assert test_group_profile_.group.name == test_group.name
    assert test_group_profile_.group.screen_name == test_group.screen_name



from mitama.server import run_app
from mitama.server.response import Response

def app(request):
    if request.method == 'POST':
        post = request.post()
        print(post['some'])
    return Response(text='''
<!DOCTYPE html>
<html>
    <body>
        <form method='POST'>
            <input type='text' name='some'>
            <button>submit</button>
        </form>
    </body>
</html>
                    ''')

run_app(app, 8080)



from mitama.http import Response
from mitama.app import Template, PackageLoader, Environment

env = Environment(loader = PackageLoader(__package__, './templates'))

async def welcome(request):
    template = env.get_template('welcome.html')
    return Response.render(template)



from mitama.app.router import Router, Route, Path
from mitama.app.method import *

def test_router():
    path_a=Path('/huga/<name>')
    path_e=Path('/huga/<id:int>')
    path_f=Path('/huga/<id:float>')
    path_a_=Path('/huga/<name>/hoge')
    path_b=Path('/hogera/<path:re:.*>')
    path_b_=Path('/hogera/<path:path>')
    assert path_a.match('/huga/boke0')
    assert path_e.match('/huga/0')
    assert path_f.match('/huga/0.1')
    assert not path_a.match('/huga/boke0/hoge')
    assert path_a_.match('/huga/boke0/hoge')
    assert path_b.match('/hogera/var/www/html') == {
        'path': 'var/www/html'
    }
    assert path_b_.match('/hogera/hogera.py') == {
        'path': 'hogera.py'
    }





from mitama.app import Controller
from mitama.http import Response

class WelcomeController(Controller):
    def handle(self, request):
        template = self.view.get_template('welcome.html')
        return Response.render(template)



#!/usr/bin/python
'''プロジェクト作成コマンド

    * mitama.json置くだけのコマンド

'''
import os
import json
from pathlib import Path
from mitama.conf import Config

def init_project_dir(path):
    with open(path / 'mitama.json', mode = 'w') as f:
        conf = Config(path, {
            'apps': {
                'mitama.portal': {
                    'path': '/'
                }
            }
        })
        data = conf.to_dict()
        json_text = json.dumps(data, indent=2)
        f.write(json_text)

class Command:
    def handle(self, argv = None):
        path = Path(os.getcwd())
        init_project_dir(path)
